/*
Copyright (c) 2012, Felix J. Ogris
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met: 

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer. 
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <limits.h>
#include <err.h>
#include <alloca.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <fcntl.h>
#include "corosync/cfg.h"
#include "corosync/cpg.h"
#include "common.h"

#define MAX_NODES 128

#define CS_ERRX(func, params ...) do { \
  if ((error = func(params)) != CS_OK) \
    errx(1, "%s() failed: %s (%i)", #func, cs_error_to_string(error), error); \
} while (0)

static int glbl_verbose = 0;
static char glbl_node_addresses[MAX_NODES][NODE_ADDRESSES_BUFFER_SIZE];
static unsigned int glbl_nodeids[MAX_NODES];
static unsigned int glbl_num_nodeids;
static char glbl_fullpath[PATH_MAX];
static unsigned int *glbl_acked_nodeids;
static struct stat glbl_stat;

void cfg_state_track_callback (
  corosync_cfg_state_notification_buffer_t *notification_buffer,
  cs_error_t error)
{
  notification_buffer = notification_buffer;
  error = error;
}

void cfg_shutdown_callback (corosync_cfg_handle_t cfg_handle,
                            corosync_cfg_shutdown_flags_t flags)
{
  cfg_handle = cfg_handle;
  flags = flags;
}

void cpg_deliver_callback (cpg_handle_t handle,
                           const struct cpg_name *group_name, uint32_t nodeid,
                           uint32_t pid, void *msg, size_t msg_len)
{
  unsigned char opcode;
  char *c;
  char *filename;
  char *error;
  unsigned int ui;
  char *addrs = "?";
  ssize_t filesize;

  handle = handle;
  pid = pid;

  if (strncmp(GROUPNAME, group_name->value, group_name->length))
    /* not for us */
    return;

  /* packet format (host byte order):
     opcode      (unsigned char)
     filename\0  (char[])
     if opcode == CS_OK: filesize    (ssize_t)
     else:               errortext\0 (char[])
   */
  if (msg_len < sizeof(opcode) + sizeof(char)) return;

  for (ui = 0; ui < glbl_num_nodeids; ui++)
    if (nodeid == glbl_nodeids[ui]) addrs =  glbl_node_addresses[ui];

  memcpy(&opcode, msg, sizeof(opcode));
  switch (opcode) {
    case OP_OK:
    case OP_ERR: break;
    case OP_PUSH: return;
    default:
      fprintf(stderr, "unknown opcode %u from node %u (%s)\n",
              opcode, nodeid, addrs);
      return;
  }

  filename = msg + sizeof(opcode);
  for (c = filename; (void*) c < msg + msg_len; c++)
    if (*c == '\0') break;

  if ((void*) c >= msg + msg_len) {
    fprintf(stderr, "message from node %u (%s) too short for filename\n",
            nodeid, addrs);
    return;
  }

  if (strcmp(filename, glbl_fullpath)) {
    fprintf(stderr, "message from node %u (%s) about file %s, not %s\n",
            nodeid, addrs, filename, glbl_fullpath);
    return;
  }

  c++;

  if (opcode == OP_OK) {
    if ((void*) c + sizeof(filesize) > msg + msg_len) {
      fprintf(stderr, "message from node %u (%s) too short for filesize\n",
              nodeid, addrs);
      return;
    }

    memcpy(&filesize, c, sizeof(filesize));

    if (filesize != glbl_stat.st_size)
      fprintf(stderr, "node %u (%s): ack'ed %lu B, sent %lu B\n",
              nodeid, addrs, filesize, glbl_stat.st_size);
    else if (glbl_verbose)
      fprintf(stdout, "node %u (%s): OK\n", nodeid, addrs);
  } else {
    error = c;
    for (; (void*) c < msg + msg_len; c++)
      if (*c == '\0') break;
    if ((void*) c >= msg + msg_len) {
      fprintf(stderr, "message from node %u (%s) too short for error text\n",
              nodeid, addrs);
      return;
    }
    fprintf(stderr, "node %u (%s): ERROR: %s\n", nodeid, addrs, error);
  }

  for (ui = 0; ui < glbl_num_nodeids; ui++) {
    if (glbl_nodeids[ui] == nodeid) {
      if (glbl_acked_nodeids[ui])
        fprintf(stderr, "duplicate response from node %u (%s)\n",
                nodeid, addrs);
      glbl_acked_nodeids[ui] = 1;
    }
  }
}

void cpg_confchg_callback (cpg_handle_t handle,
                           const struct cpg_name *group_name,
                           const struct cpg_address *member_list,
                           size_t member_list_entries,
                           const struct cpg_address *left_list,
                           size_t left_list_entries,
                           const struct cpg_address *joined_list,
                           size_t joined_list_entries)
{
  handle = handle;
  group_name = group_name;
  member_list = member_list;
  member_list_entries = member_list_entries;
  left_list = left_list;
  left_list_entries = left_list_entries;
  joined_list = joined_list;
  joined_list_entries = joined_list_entries;
}

/* returns current dir + filename, if filename doesn't start w/ '/'
   otherwise filename itself is returned */
void make_fullpath (char *filename, char *fullpath)
{
  if (*filename == '/') fullpath[0] = '\0';
  else {
    if (getcwd(fullpath, PATH_MAX) == NULL)
      err(1, "cannot determine current directory");
    strncat(fullpath, "/", PATH_MAX - strlen(fullpath) - sizeof(char));
  }

  strncat(fullpath, filename, PATH_MAX - strlen(fullpath) - sizeof(char));
}

void send_file (cpg_handle_t cpg_handle, char *fullpath, struct stat *st)
{
  int fd;
  struct iovec iov[3];
  cs_error_t error;
  struct header hdr;
  int iovcnt;

  if ((fd = open(fullpath, O_RDONLY)) < 0)
    err(1, "cannot open %s for reading", fullpath);

  /* packet format (host byte order):
     opcode      (unsigned char)
     filemode    (mode_t)
     owner       (uid_t)
     group       (gid_t)
     fullpath\0  (char[])
     data        (void[])
   */
  hdr.opcode = OP_PUSH;
  hdr.mode = st->st_mode;
  hdr.uid = st->st_uid;
  hdr.gid = st->st_gid;
  
  iov[0].iov_base = &hdr;
  iov[0].iov_len = sizeof(hdr);
  iov[1].iov_base = fullpath;
  iov[1].iov_len = strlen(fullpath) + sizeof(char);
  iov[2].iov_len = st->st_size;
  if (st->st_size > 0) {
    iov[2].iov_base = mmap(NULL, st->st_size, PROT_READ, MAP_PRIVATE, fd, 0);
    if (iov[2].iov_base == MAP_FAILED) err(1, "cannot mmap %s", fullpath);
    iovcnt = 3;
  } else iovcnt = 2;

  CS_ERRX(cpg_mcast_joined, cpg_handle, CPG_TYPE_AGREED, iov, iovcnt);

  if ((st->st_size > 0) && (munmap(iov[2].iov_base, st->st_size) < 0))
    err(1, "cannot unmap %s", fullpath);

  if (close(fd) < 0) err(1, "cannot close %s", fullpath);
}

/* wait for action on corosync socket */
void wait_nodes (cpg_handle_t cpg_handle, unsigned int *acked_nodeids,
                 unsigned int num_nodeids)
{
  cs_error_t error;
  unsigned int ui;
  int fd;
  fd_set fds;
  struct timeval timeout;
  int res;

  CS_ERRX(cpg_fd_get, cpg_handle, &fd);

  for (ui = 0; ui < num_nodeids; ui++) acked_nodeids[ui] = 0;

  for (;;) {
    for (ui = 0; ui < num_nodeids; ui++)
      if (!acked_nodeids[ui]) break;

    if (ui >= num_nodeids) return;

    FD_ZERO(&fds);
    FD_SET(fd, &fds);
    timeout.tv_sec = 60;
    timeout.tv_usec = 0;

    res = select(fd + 1, &fds, NULL, NULL, &timeout);
    if (res < 0) errx(1, "select()");
    if (res == 0) err(1, "timeout while waiting for response from nodes");
    if ((res > 0) && FD_ISSET(fd, &fds))
      CS_ERRX(cpg_dispatch, cpg_handle, CS_DISPATCH_ALL);
  }
}

void print_usage (FILE *stream)
{
  fputs("clucp " VERSION "\n"
"usage: clucp [-v] <file> [ ... ]\n"
"       clucp -l\n"
"       clucp -h\n\n"
"  <file>    file(s) to distribute among the cluster\n"
"  -v        be verbose\n"
"  -l        just list all cluster members\n"
"  -h        this help ;-)\n", stream);
}

int main (int argc, char **argv)
{
  cs_error_t error;
  corosync_cfg_handle_t cfg_handle;
  cpg_handle_t cpg_handle;
  corosync_cfg_callbacks_t cfg_callbacks = {
    .corosync_cfg_state_track_callback = cfg_state_track_callback,
    .corosync_cfg_shutdown_callback = cfg_shutdown_callback,
  };
  cpg_callbacks_t cpg_callbacks = {
    .cpg_deliver_fn = cpg_deliver_callback,
    .cpg_confchg_fn = cpg_confchg_callback,
  };
  struct cpg_name group_name = {
    .length = strlen(GROUPNAME),
    .value = GROUPNAME,
  };
  unsigned int local_nodeid;
  cpg_iteration_handle_t iter_handle;
  struct cpg_iteration_description_t iter_desc;
  unsigned int ui;
  int i;
  int list_only = 0;
  char node_addresses[NODE_ADDRESSES_BUFFER_SIZE];
  int is_local_node;

  while ((i = getopt(argc, argv, "hlv")) != -1) {
    switch (i) {
      case 'h': print_usage(stdout); return 0;
      case 'l': list_only = 1; break;
      case 'v': glbl_verbose = 1; break;
      default: fputs("for help use -h\n", stderr); return 1;
    }
  }

  if (!list_only && (optind >= argc)) {
    print_usage(stderr);
    return 1;
  }

  CS_ERRX(corosync_cfg_initialize, &cfg_handle, &cfg_callbacks);
  CS_ERRX(cpg_initialize, &cpg_handle, &cpg_callbacks);
  CS_ERRX(cpg_join, cpg_handle, &group_name);
  CS_ERRX(corosync_cfg_local_get, cfg_handle, &local_nodeid);
  CS_ERRX(cpg_iteration_initialize, cpg_handle, CPG_ITERATION_ALL,
          NULL, &iter_handle);

  glbl_num_nodeids = 0;

  if (list_only) puts("I\tgroup\tnode id\t\tpid\taddrs");

  while ((cpg_iteration_next(iter_handle, &iter_desc)) == CS_OK) {
    if (strcmp(iter_desc.group.value, GROUPNAME))
      /* not our group */
      continue;

    is_local_node = (iter_desc.nodeid == local_nodeid);
    if ((iter_desc.pid == (unsigned) getpid()) && is_local_node)
      /* this instance of clucp */
      continue;

    make_node_addresses(cfg_handle, iter_desc.nodeid, node_addresses,
                        sizeof(node_addresses));

    if (list_only)
      printf("%c\t%s\t%u\t%u\t%s\n",
             (is_local_node ? '*' : ' '), iter_desc.group.value,
             iter_desc.nodeid, iter_desc.pid, node_addresses);

    if (is_local_node)
      /* don't wait for ack from local node */
      continue;

    if (glbl_num_nodeids >= sizeof(glbl_nodeids)/sizeof(glbl_nodeids[0]))
      errx(1, "too many nodes");

    for (ui = 0; ui < glbl_num_nodeids; ui++)
      if (glbl_nodeids[ui] == iter_desc.nodeid) break;

    if (ui >= glbl_num_nodeids) {
      glbl_nodeids[glbl_num_nodeids] = iter_desc.nodeid;
      strncpy(glbl_node_addresses[ui], node_addresses,
              sizeof(glbl_node_addresses[ui]));
      glbl_num_nodeids++;
    }
  }

  CS_ERRX(cpg_iteration_finalize, iter_handle);

  if (!list_only) {
    glbl_acked_nodeids = alloca(glbl_num_nodeids * sizeof(unsigned int));

    for (i = optind; i < argc; i++) {
      make_fullpath(argv[i], glbl_fullpath);

      if (stat(glbl_fullpath, &glbl_stat) < 0)
        err(1, "cannot stat %s", glbl_fullpath);

      if (glbl_verbose)
        printf("sending %s (%li B) to %u node%s\n",
               glbl_fullpath, glbl_stat.st_size, glbl_num_nodeids,
               (glbl_num_nodeids == 1 ? "" : "s"));

      send_file(cpg_handle, glbl_fullpath, &glbl_stat);
      wait_nodes(cpg_handle, glbl_acked_nodeids, glbl_num_nodeids);
    }
  }

  CS_ERRX(cpg_leave, cpg_handle, &group_name);
  CS_ERRX(cpg_finalize, cpg_handle);
  CS_ERRX(corosync_cfg_finalize, cfg_handle);

  return 0;
}
